#!/usr/bin/lua

-- wbgen2, (c) 2010 Tomasz Wlostowski/CERN BE-Co-HT
-- LICENSED UNDER GPL v2

-- File: cgen_vhdl.lua
-- The VHDL code generator

fieldtype_2_vhdl={};
fieldtype_2_vhdl[BIT]="std_logic";
fieldtype_2_vhdl[MONOSTABLE]="std_logic";
fieldtype_2_vhdl[SIGNED] = "signed";
fieldtype_2_vhdl[UNSIGNED] = "unsigned";
fieldtype_2_vhdl[ENUM] = "std_logic_vector";
fieldtype_2_vhdl[SLV] = "std_logic_vector";

cvtfunc_hdl = {};
cvtfunc_hdl[SLV] = "";
cvtfunc_hdl[SIGNED] = "signed";
cvtfunc_hdl[UNSIGNED] = "unsigned";

-- generates a string containing VHDL-compatible binary numeric value of size numbits
function gen_vhdl_bin_literal(value, numbits)
    local str ='\"';
    local i,n,d,r;
    
    n=value;
    r=math.pow(2, numbits-1);
    
    for i=1,numbits do
	d=math.floor(n/r);
	str=str..csel(d>0,"1","0");
	n=n%r;
	r=r/2;
    end
    return str..'\"';
end


function cgen_vhdl_header()
    emit("---------------------------------------------------------------------------------------");
    emit("-- Title          : Wishbone slave core for "..periph.name);
    emit("---------------------------------------------------------------------------------------");
    emit("-- File           : "..options.output_hdl_file);
    emit("-- Author         : auto-generated by wbgen2 from "..input_wb_file);
    emit("-- Created        : "..os.date());
    emit("-- Standard       : VHDL'87");
    emit("---------------------------------------------------------------------------------------");
    emit("-- THIS FILE WAS GENERATED BY wbgen2 FROM SOURCE FILE "..input_wb_file);
    emit("-- DO NOT HAND-EDIT UNLESS IT'S ABSOLUTELY NECESSARY!");
    emit("---------------------------------------------------------------------------------------");
		emit("");
		emit("library ieee;");
		emit("use ieee.std_logic_1164.all;");
		emit("use ieee.numeric_std.all;");
		emit("");
end

function cgen_vhdl_entity()
	local last;

	emit ("entity "..periph.hdl_entity.." is");
	indent_right();
  emit ("port (");
	indent_right();

  for i=1,table.getn(g_portlist) do
		local port = g_portlist[i];
		  
		if(i == table.getn(g_portlist)) then
			last = true;
		else
		  last = false;
		end

		if(port.comment ~= nil) then
			emitx("-- "..port.comment.."\n");
		end
	
		local line = string.format("%-40s : %-6s %s", port.name, port.dir, fieldtype_2_vhdl[port.type]);

		if(port.range > 0) then
			line = line.."("..(port.range-1).." downto 0)";
		end    
		      
		line=line..csel(last, "", ";");
		emit(line);
  end

    
	indent_left();
 	emit(");");
	indent_left();
	emit("end "..periph.hdl_entity..";");
	emit("");
	emit("architecture syn of "..periph.hdl_entity.." is");
	emit("");

	for i,v in pairs (g_siglist) do
		s=string.format("signal %-40s : %-15s", v.name, fieldtype_2_vhdl[v.type]);
		if(v.range > 0) then
			s=s..string.format("(%d downto 0)", v.range-1);
		end
		s=s..";";
		emit(s);
	end
	emit("");
	emit("begin");
	indent_right();
	
end

function cgen_vhdl_ending()
	indent_left();
	emit("end syn;");
end

-- Generates the VHDL code from syntax tree <tree>, writing it into previously chosen file.
function cgen_generate_vhdl_code(tree)

function find_code(node, t)
	for i,v in ipairs(node) do if ((v.t ~= nil) and (v.t == t)) then return v; end end
	return nil;
end

function cgen_vhdl_syncprocess(node)
	emit("process ("..node.clk..", "..node.rst..")");
	emit("begin");
	indent_right();
			
	local vrst = find_code(node.code, "reset");
	local vpe = find_code(node.code, "posedge");
			
	if(vpe == nil) then die("vhdl code generation error: no vposedge defined for vsyncprocess"); end
			
	if(options.reset_type == "asynchronous") then
			
		if(vrst ~= nil) then
			emit("if ("..node.rst.." = '"..vrst.level.."') then ");
			indent_right();
			recurse(vrst.code);
			indent_left();
	
			emit("elsif rising_edge("..node.clk..") then");
			indent_right();
		
		else
			emit("if rising_edge("..node.clk..") then");			
			indent_right();
		end

		recurse(vpe.code);		
	  indent_left();
 	  emit("end if;");
			
	else -- sync reset
		emit("if rising_edge("..node.clk..") then");
	  indent_right();
			
		if(vrst ~= nil) then
			 emit("if ("..node.rst.." = '"..vrst.level.."') then ");
			 indent_right();
			 recurse(vrst.code);
			 indent_left();
			 emit("else ");
		end
 		  
	  indent_right();
		recurse(vpe.code); 		  
	  indent_left();
	  emit("end if;");
	  indent_left();
	  emit("end if;");
	end
	indent_left();

	emit("end process;");	
	emit("");
end

function node_typesize(node)
	local ts = {};
	local sig;

	print(node);

	if(type(node) == "table") then
		print("table!");
		if(node.t ~= nil and node.t == "index") then
				sig = cgen_find_sigport(node.name);

				ts.h=node.h;
				ts.l=node.l;
				ts.name=sig.name;
				ts.type=sig.type;
				ts.size = csel(ts.l==nil,1,ts.h-ts.l+1);
			
				return ts;			
	 	else
				print("expr2: ", node.t);
				ts.type = EXPRESSION;
				ts.code = node;
				return ts;
		end

	 elseif(type(node) == "string") then
		print("string-type node: "..node);
		sig = cgen_find_sigport(node);
		ts.size = sig.range;
		ts.type = sig.type;
		ts.name = node;
		return ts;
	 elseif(type(node) == "number") then
		ts.type = INTEGER;
		ts.name = node;
		ts.size = 0;
		return ts;
	else 
		print("Unknown node?");
end
end

function gen_subrange(t)
	if(t.h ~= nil and t.l == nil) then
		return t.name.."("..t.h..")";
	elseif(t.h ~= nil and t.l ~= nil) then
		return t.name.."("..t.h.." downto "..t.l..")";
	else
		return t.name;
	end
end

function calc_size(t)
	if(t.h ~= nil and t.l == nil) then
		return 1;
	elseif(t.h ~= nil and t.l ~= nil) then
		return t.h-t.l+1; 
	else
		local sig= cgen_find_sigport(t.name);
		return sig.range;
	end
end

function gen_vhdl_typecvt(tsd, tss)
	if(tsd.type == tss.type) then -- assign same types
		return(gen_subrange(tss));
	elseif (tss.type == INTEGER) then
		if(tsd.type == BIT) then -- assign bit/slv/signed/unsigned <= number;
			return("'"..tss.name.."'");
		elseif(tsd.type == SLV) then
			return("std_logic_vector(to_unsigned("..tss.name..", "..calc_size(tsd).."))");
		elseif(tsd.type == SIGNED) then
			return("to_signed("..tss.name..", "..calc_size(tsd)..")");
		elseif(tsd.type == UNSIGNED) then
			return("to_unsigned("..tss.name..", "..calc_size(tsd)..")");
		else die ("unsupported assignment: "..tsd.name.." "..tss.name); end	
	elseif (tss.type == BIT) then
		if(tsd.type == SLV) then
			return(gen_subrange(tss.name));
		else die ("unsupported assignment: "..tsd.name.." "..tss.name); end	
	elseif (tss.type == SIGNED or tss.type == UNSIGNED) then
		if(tsd.type == SLV) then
			return("std_logic_vector("..gen_subrange(tss.name)..")");
		else die ("unsupported assignment: "..tsd.name.." "..tss.name); end	
	elseif (tss.type == SLV) then
		if(tsd.type == SIGNED) then
			return("signed("..gen_subrange(tss.name)..")");
		elseif (tsd.type == UNSIGNED) then
			return("unsigned("..gen_subrange(tss.name)..")");
		else die ("unsupported assignment: "..tsd.name.." "..tss.name); end	
	else die ("unsupported assignment: "..tsd.name.." "..tss.name); end	
end

function cgen_vhdl_assign(node)
	local tsd = node_typesize(node.dst);
	local tss = node_typesize(node.src);

	for i,v in pairs(tsd) do
	print(i,v);
	end

	if(tss.type == EXPRESSION) then
		print("EXPR!");
		emiti();
		emitx(gen_subrange(tsd).." <= ");
		recurse({tss.code});
		emitx(";\n");
	else
		emit(gen_subrange(tsd).." <= "..gen_vhdl_typecvt(tsd, tss)..";");
	end

end

function cgen_vhdl_if(node)
	emiti(); emitx("if (");
	print("EMIT IF:", node.cond);

	for i,v in pairs(node.cond) do print(i,v); end
	recurse(node.cond);
	emitx(") then\n");
	if(node.code_else ~= nil) then
	indent_right();
	recurse(node.code);
	indent_left();
	emit("else");
	indent_right();
	recurse(node.code_else);
	indent_left();
	emit("end if;");
	else
	indent_right();
	recurse(node.code);
	indent_left();
	emit("end if;");

	end
	
end

function cgen_vhdl_not(node)
	local tsa = node_typesize(node.a);

	emitx("not ");

	if(tsa.type == EXPRESSION) then
		emitx("("); recurse({node.a}); emitx(")");
	else
		emitx(gen_subrange(tsa));
	end
end


function cgen_vhdl_op(node, op)
	local tsa = node_typesize(node.a);
	local tsb = node_typesize(node.b);

	if(tsa.type == EXPRESSION) then
		emitx("("); recurse({node.a}); emitx(")");
	else
		emitx(gen_subrange(tsa));
	end

	if(op=="eq") then emitx(" = "); end
	if(op=="and") then emitx(" and "); end
	if(op=="or") then emitx(" or "); end
	if(op=="sub") then emitx(" - "); end
	if(op=="add") then emitx(" + "); end

	if(tsb.type == EXPRESSION) then
		emitx("("); recurse({node.b}); emitx(")");
	else
		emitx(gen_vhdl_typecvt(tsa, tsb));
	end
end


function cgen_vhdl_comment(node)
	emitx("-- "..node.str.."\n");
end


-- switch/case statement --
function cgen_vhdl_switch(node)

	local tsa = node_typesize(node.a);

	emiti(); emitx("case ");

	if(tsa.type == EXPRESSION) then
		emitx("("); recurse({node.a}); emitx(")");
	else
		local tsb = {};
		tsb.type = SLV;
		emitx(gen_vhdl_typecvt(tsb, tsa));
	end	

	emitx(" is\n");

	for i,v in pairs(node.code)  do
		print("CA: ", v.t);

		if(v.t == "case") then			
			emit("when "..gen_vhdl_bin_literal(v.a, tsa.size).." => ");
			indent_right();
			recurse({v.code});
			indent_left();
		elseif(v.t == "casedefault") then			
			emit("when others =>");
			indent_right();
			recurse({v.code});
			indent_left();
		end	


	end

	emit("end case;");

end



function recurse(t)


	print("recurse: ", t);

	for i,v in pairs(t) do
		print("-> i ",i,"v ",v, "vt ",v.t);

		if(v.t == nil) then
			recurse(v);
		elseif(v.t == "comment") then
			cgen_vhdl_comment(v);
		elseif(v.t == "syncprocess") then
		 cgen_vhdl_syncprocess(v);
		elseif(v.t=="assign") then
		 cgen_vhdl_assign(v);		
		elseif(v.t=="if") then
			cgen_vhdl_if(v);
		elseif(v.t=="eq" or v.t=="sub" or v.t=="add" or v.t == "or" or v.t=="and") then
			cgen_vhdl_op(v,v.t);
		elseif(v.t=="switch") then
			cgen_vhdl_switch(v);
		elseif(v.t=="not") then
			cgen_vhdl_not(v);
		else
			print("unimpl: "..v.t);
			os.exit(-1);	
		end
	end
	print("end recurse");
end


cgen_new_snippet();
cgen_vhdl_header();
cgen_vhdl_entity();
recurse(tree);
cgen_vhdl_ending();
cgen_write_current_snippet();
end
